"""Extractor class file."""
# *modified "Fri Oct 16 16:27:41 2015" *by "Paul E. Black"

import re
import os
import lxml.etree as ET
import fnmatch
from sets import Set
from ocse import site
import threading
import Queue
from pycparser import parse_file, c_ast, c_generator
from ocse.node_visitor import *

#Path to the c preprocessor
CPPPATH = 'cpp'

class Extractor:

  """Class to extract sites from Juliet 1.2 and check if these sites are buggy
  or not.
  """

  def __init__(self,
      root_path,
      parse_single_cwe=None):
    """root_path is the path where the juliet files to be analyzed are.
    This constructor method prepares all the data structures to receive
    the Synthetic Trees informations from pycparser.
    parse_single_cwe makes ocse run faster by getting results for a
    single cwe. current supported options are:
    'division_by_zero'
    'variable_access'
    'null_ptr'
    'int_of'
    'int_uf'
    'buffer_write'
    'buffer_read'
    'write_what_where'
    'return'
    """
    self.root_path = root_path
    self.parse_single_cwe = parse_single_cwe
    self.files = []
    self.set_files_list()
    self.ast_queue = Queue.Queue()
    self.ast_queue_done = False
    self.ast_divisions = Queue.Queue()
    self.ast_variable_accesses = Queue.Queue()
    self.ast_null_ptr_dereferences = Queue.Queue()
    self.ast_int_overflows = Queue.Queue()
    self.ast_int_underflows = Queue.Queue()
    self.ast_buffer_writes = Queue.Queue()
    self.ast_buffer_reads = Queue.Queue()
    self.ast_write_what_wheres = Queue.Queue()
    self.ast_returns = Queue.Queue()
    self.lines_map = {}
    self.extract()

  def extract(self):
    """Trigger the threads to generate the ATSs and navigate through them."""
    extract_th = threading.Thread(None, self.extract_ast())
    populate_th = threading.Thread(None, self.populate_ast_attributes())
    extract_th.start()
    populate_th.start()
    extract_th.join()
    populate_th.join()
    self.map_lines()

  def set_files_list(self):
    """Navigates through the filepath tree and appends all C files in the files
    list.
    """
    for root, dirnames, filenames in os.walk(self.root_path):
      for filename in fnmatch.filter(filenames, '*.c'):
        self.files.append(os.path.join(root, filename))

  def map_lines(self):
    """Get the code in the lines pointed by the sites in the Abstract Syntax
    Trees and append them to lines_map with the line number (code:line).
    """
    ast_lists = list(self.ast_divisions.queue) + list(self.ast_variable_accesses.queue) + list(self.ast_null_ptr_dereferences.queue) + list(self.ast_int_overflows.queue) + list(self.ast_int_underflows.queue) + list(self.ast_buffer_writes.queue) + list(self.ast_buffer_reads.queue) + list(self.ast_write_what_wheres.queue) + list(self.ast_returns.queue)

    ast_path_map = {}
    for ast_object in ast_lists:
      if(not ast_path_map.has_key(ast_object.coord.file)):
        ast_path_map[ast_object.coord.file] = []
      ast_path_map[ast_object.coord.file].append(ast_object)

    for file_path, ast_objects in ast_path_map.iteritems():
      f = open(file_path, 'r')
      lines = f.readlines()
      for ast_object in ast_objects:
        site_line = re.sub('(  |\r|\n|\t)',"",lines[ast_object.coord.line - 1])
        self.lines_map["{0}:{1}".format(file_path, ast_object.coord.line)] = site_line

      f.close

  def variable_access_sites(self):
    """Returns list of Variable Access sites."""
    sites = []
    while(not self.ast_variable_accesses.empty()):
      variable_access = self.ast_variable_accesses.get()
      site_line = self.lines_map["{0}:{1}".format(variable_access.coord.file, variable_access.coord.line)]
      info = (c_generator.CGenerator()).visit(variable_access)
      sites.append(site.Site(variable_access.coord.file, "variable", variable_access.coord.line, site_line, info))

    Extractor.set_buggy_attr(sites, self.buggy_sites())
    return sites

  def div_by_zero_sites(self):
    """Returns list of Division by Zero sites."""
    sites = []
    while(not self.ast_divisions.empty()):
      division = self.ast_divisions.get()
      site_line = self.lines_map["{0}:{1}".format(division.coord.file, division.coord.line)]
      if(isinstance(division, c_ast.Assignment)):
        # exclude instances of var /= constant
        if(isinstance(division.rvalue, c_ast.Constant)):
          continue

        info = (c_generator.CGenerator()).visit(division.rvalue)

      elif(isinstance(division, c_ast.BinaryOp)):
        # exclude instances of var / constant
        if(isinstance(division.right, c_ast.Constant)):
          continue

        info = (c_generator.CGenerator()).visit(division.right)

      sites.append(site.Site(division.coord.file, "div_by_zero", division.coord.line, site_line, info))

    Extractor.set_buggy_attr(sites, self.buggy_sites())
    return sites

  def null_ptr_deref_sites(self):
    """Returns list of Pointer Dereference sites."""
    sites = []
    while(not self.ast_null_ptr_dereferences.empty()):
      null_ptr_deref = self.ast_null_ptr_dereferences.get()
      site_line = self.lines_map["{0}:{1}".format(null_ptr_deref.coord.file, null_ptr_deref.coord.line)]
      info = (c_generator.CGenerator()).visit(null_ptr_deref)
      sites.append(site.Site(null_ptr_deref.coord.file, "null_ptr_deref", null_ptr_deref.coord.line, site_line, info))

    Extractor.set_buggy_attr(sites, self.buggy_sites())
    return sites

  def int_overflow_sites(self):
    """Returns list of Integer overflow sites."""
    sites = []
    while(not self.ast_int_overflows.empty()):
      int_overflow = self.ast_int_overflows.get()
      site_line = self.lines_map["{0}:{1}".format(int_overflow.coord.file, int_overflow.coord.line)]
      info = (c_generator.CGenerator()).visit(int_overflow)
      sites.append(site.Site(int_overflow.coord.file, "int_overflow", int_overflow.coord.line, site_line, info))

    Extractor.set_buggy_attr(sites, self.buggy_sites())
    return sites

  def int_underflow_sites(self):
    """Returns list of Integer underflow sites."""
    sites = []
    while(not self.ast_int_underflows.empty()):
      int_underflow = self.ast_int_underflows.get()
      site_line = self.lines_map["{0}:{1}".format(int_underflow.coord.file, int_underflow.coord.line)]
      info = (c_generator.CGenerator()).visit(int_underflow)
      sites.append(site.Site(int_underflow.coord.file, "int_underflow", int_underflow.coord.line, site_line, info))

    Extractor.set_buggy_attr(sites, self.buggy_sites())
    return sites

  def buffer_write_sites(self):
    """Returns list of buffer write sites.
    Buffer write sites are stack and heap based buffer overflows and buffer underwrites.
    """
    sites = []
    while(not self.ast_buffer_writes.empty()):
      buffer_write = self.ast_buffer_writes.get()
      site_line = self.lines_map["{0}:{1}".format(buffer_write.coord.file, buffer_write.coord.line)]
      info = (c_generator.CGenerator()).visit(buffer_write)
      sites.append(site.Site(buffer_write.coord.file, "buffer_write", buffer_write.coord.line, site_line, info))

    Extractor.set_buggy_attr(sites, self.buggy_sites())
    return sites

  def buffer_read_sites(self):
    """Returns list of buffer read sites.
    Buffer read sites are buffer under reads and buffer over reads candidates.
    """
    sites = []
    while(not self.ast_buffer_reads.empty()):
      buffer_read = self.ast_buffer_reads.get()
      site_line = self.lines_map["{0}:{1}".format(buffer_read.coord.file, buffer_read.coord.line)]
      info = (c_generator.CGenerator()).visit(buffer_read)
      sites.append(site.Site(buffer_read.coord.file, "buffer_read", buffer_read.coord.line, site_line, info))

    Extractor.set_buggy_attr(sites, self.buggy_sites())
    return sites

  def write_what_where_sites(self):
    """Returns list of Write What Where sites."""
    sites = []
    while(not self.ast_write_what_wheres.empty()):
      write_what_where = self.ast_write_what_wheres.get()
      site_line = self.lines_map["{0}:{1}".format(write_what_where.coord.file, write_what_where.coord.line)]
      info = (c_generator.CGenerator()).visit(write_what_where)
      sites.append(site.Site(write_what_where.coord.file, "write_what_where", write_what_where.coord.line, site_line, info))

    Extractor.set_buggy_attr(sites, self.buggy_sites())
    return sites

  def return_sites(self):
    """Returns list of return sites."""
    sites = []
    while(not self.ast_returns.empty()):
      returned = self.ast_returns.get()
      site_line = self.lines_map["{0}:{1}".format(returned.coord.file, returned.coord.line)]
      info = (c_generator.CGenerator()).visit(returned)
      sites.append(site.Site(returned.coord.file, "return", returned.coord.line, site_line, info))

    Extractor.set_buggy_attr(sites, self.buggy_sites())
    return sites

  def buggy_sites(self):
    """Parses all files in the filepath looking for buggy sites according to
    specific rules for each type of site.
    """
    sites = []
    for file_path in self.files:
      f = open(file_path, 'r')
      inside_bad = False
      lines = f.readlines()
      for line_number, line in enumerate(lines):
        if(re.search("#ifndef OMITBAD", line) != None):
          inside_bad = True
          continue

        elif(re.search("#endif /\* OMITBAD \*/", line) != None):
          inside_bad = False
          continue

        elif(inside_bad):
          #div_zero
          if(re.search("POTENTIAL FLAW: Possibly divide by zero", lines[line_number-1]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "div_by_zero", line_number + 1, site_line, None, True))

          #variable_access
          elif(re.search("POTENTIAL FLAW: Use data without initializing it", lines[line_number-1]) != None or
              re.search("POTENTIAL FLAW: Use data without initializing it", lines[line_number-2]) != None or
              re.search("POTENTIAL FLAW: Use data without initializing it", lines[line_number-5]) != None or
              re.search("POTENTIAL FLAW: Use data without initializing it", lines[line_number-6]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "variable", line_number + 1, site_line, None, True))

          #null_ptr_deref
          elif(re.search("thus causing a NPD", lines[line_number-1]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "null_ptr_deref", line_number + 1, site_line, None, True))

          elif(re.search("POTENTIAL FLAW: Attempt to use data, which may be NULL", lines[line_number-1]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "null_ptr_deref", line_number + 1, site_line, None, True))

          elif(re.search("FLAW: Check for NULL, but still dereference the pointer", lines[line_number-1]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", lines[line_number + 3])
            sites.append(site.Site(file_path, "null_ptr_deref", line_number + 4, site_line, None, True))

          elif(re.search("checks for NULL, so we cannot use it here", lines[line_number-1]) != None):
            if(re.search("POTENTIAL FLAW: Attempt to use data, which may be NULL", lines[line_number-2]) != None):
              site_line = re.sub('(  |\r|\n|\t)', "", line)
              sites.append(site.Site(file_path, "null_ptr_deref", line_number + 1, site_line, None, True))

          elif(re.search('intPointer = \(int \*\)malloc\(', lines[line_number-1]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "null_ptr_deref", line_number + 1, site_line, None, True))

          #int_overflow
          elif(re.search("Adding 1 to data could cause an", lines[line_number-1]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "int_overflow", line_number + 1, site_line, None, True))

          elif(re.search(", this will overflow", lines[line_number-1]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "int_overflow", line_number + 1, site_line, None, True))

          #int_underflow
          elif(re.search("Subtracting 1 from data could cause an underflow", lines[line_number-1]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "int_underflow", line_number + 1, site_line, None, True))

          elif(re.search(", this will underflow", lines[line_number-1]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "int_underflow", line_number + 1, site_line, None, True))

          #buffer overflow (CWE121, 122, 124)
          #stack based buffer overflow files
# Don't count these as buggy.  These are associated with      PEB 28 Sept 2015
#        memcpy(structCharVoid.charFirst, SRC_STR, sizeof(structCharVoid));
# C standard says overwriting fields in structs is ok
#          elif(re.search("FLAW: Use the sizeof\(structCharVoid\) which will overwrite the pointer voidSecond", lines[line_number-1]) != None):
#            site_line = re.sub('(  |\r|\n|\t)', "", line)
#            sites.append(site.Site(file_path, "buffer_write", line_number + 1, site_line, None, True))

          elif(re.search("POTENTIAL FLAW: Attempt to write to an index of the array that is above the upper bound", lines[line_number-4]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "buffer_write", line_number + 1, site_line, None, True))

          elif(re.search("POTENTIAL FLAW: Possible buffer overflow if data was not allocated correctly in the source", lines[line_number-1]) != None or
              re.search("POTENTIAL FLAW: Possible buffer overflow if data was not allocated correctly in the source", lines[line_number-3]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "buffer_write", line_number + 1, site_line, None, True))

          elif(re.search("POTENTIAL FLAW: treating pointer as a char\* when it may point to a wide string", lines[line_number-3]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "buffer_write", line_number + 1, site_line, None, True))

          elif(re.search("POTENTIAL FLAW: data may not have enough space to hold source", lines[line_number-1]) != None or
              re.search("POTENTIAL FLAW: data may not have enough space to hold source", lines[line_number-3]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "buffer_write", line_number + 1, site_line, None, True))

          elif(re.search("POTENTIAL FLAW: Possible buffer overflow if the size of data is less than the length of source", lines[line_number-1]) != None or
              re.search("POTENTIAL FLAW: Possible buffer overflow if the size of data is less than the length of source", lines[line_number-3]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "buffer_write", line_number + 1, site_line, None, True))

          elif(re.search("POTENTIAL FLAW: Possible buffer overflow if the sizeof\(data\)-strlen\(data\) is less than the length of source", lines[line_number-1]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "buffer_write", line_number + 1, site_line, None, True))

          elif(re.search("POTENTIAL FLAW: Possible buffer overflow if data < 100", lines[line_number-1]) != None or
              re.search("POTENTIAL FLAW: Possible buffer overflow if data < 100", lines[line_number-3]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "buffer_write", line_number + 1, site_line, None, True))

          elif(re.search("POTENTIAL FLAW: Possible buffer overflow if data is larger than dest", lines[line_number-1]) != None or
            re.search("POTENTIAL FLAW: Possible buffer overflow if data is larger than dest", lines[line_number-3]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "buffer_write", line_number + 1, site_line, None, True))

          elif(re.search("POTENTIAL FLAW: Possible buffer overflow if data is larger than sizeof\(dest\)-strlen\(dest\)", lines[line_number-1]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "buffer_write", line_number + 1, site_line, None, True))

          elif(re.search("Possible buffer overflow if data is larger than sizeof\(dest\)-wcslen\(dest\)", lines[line_number-1]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "buffer_write", line_number + 1, site_line, None, True))

          #heap based buffer overflow files
# Don't count these as buggy.  These are associated with      PEB 28 Sept 2015
#        memcpy(structCharVoid->charFirst, SRC_STR, sizeof(*structCharVoid));
# C standard says overwriting fields in structs is ok
#          elif(re.search("FLAW: Use the sizeof\(\*structCharVoid\) which will overwrite the pointer y", lines[line_number-1]) != None):
#            site_line = re.sub('(  |\r|\n|\t)', "", line)
#            sites.append(site.Site(file_path, "buffer_write", line_number + 1, site_line, None, True))

          elif(re.search("POTENTIAL FLAW: Possible buffer overflow if source is larger than data", lines[line_number-1]) != None or
              re.search("POTENTIAL FLAW: Possible buffer overflow if source is larger than data", lines[line_number-3]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "buffer_write", line_number + 1, site_line, None, True))

          elif(re.search("POTENTIAL FLAW: Possible buffer overflow if source is larger than sizeof\(data\)-strlen\(data\)", lines[line_number-1]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "buffer_write", line_number + 1, site_line, None, True))

          elif(re.search("FLAW: Using sizeof the pointer and not the data type in malloc\(\)", lines[line_number-2]) != None):
# This makes exactly two sites buggy, when they're not          PEB 30 Sep 2015
#               or re.search("FLAW: Using sizeof the pointer and not the data type in malloc\(\)", lines[line_number-3]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "buffer_write", line_number + 1, site_line, None, True))

          #buffer underwrite files
          elif(re.search("POTENTIAL FLAW: Possibly copying data to memory before the destination buffer", lines[line_number-1]) != None or
               re.search("POTENTIAL FLAW: Possibly copying data to memory before the destination buffer", lines[line_number-3]) != None):
            if(re.search("Ensure the destination buffer is null terminated", lines[line_number-1]) == None):
              site_line = re.sub('(  |\r|\n|\t)', "", line)
              sites.append(site.Site(file_path, "buffer_write", line_number + 1, site_line, None, True))

          elif(re.search("POTENTIAL FLAW: Attempt to access a negative index of the array", lines[line_number-4]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            if(re.search("CWE124_Buffer_Underwrite", file_path) != None):
              sites.append(site.Site(file_path, "buffer_write", line_number + 1, site_line, None, True))

            #this could also be a buffer underread bug
            elif(re.search("CWE127_Buffer_Underread", file_path) != None):
              sites.append(site.Site(file_path, "buffer_read", line_number + 1, site_line, None, True))

          #buffer overread files
          elif(re.search("POTENTIAL FLAW: using length of the dest where data", lines[line_number-4]) != None and
            #re.search("printLine(dest);", lines[line_number]) == None and
            re.search("could be smaller than dest causing buffer overread", lines[line_number-3]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "buffer_read", line_number + 1, site_line, None, True))

          elif(re.search("POTENTIAL FLAW: using (memcpy|memmove) with the length of the dest where data", lines[line_number-2]) != None and
            re.search("could be smaller than dest causing buffer overread", lines[line_number-1]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "buffer_read", line_number + 1, site_line, None, True))

          elif(re.search("POTENTIAL FLAW: Attempt to access an index of the array that is above the upper bound", lines[line_number-4]) != None and
            re.search("This check does not check the upper bounds of the array index", lines[line_number-3]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "buffer_read", line_number + 1, site_line, None, True))

          elif(re.search("FLAW: do not explicitly null terminate dest", lines[line_number-1]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "buffer_read", line_number + 1, site_line, None, True))

          #buffer underread files
          elif(re.search("POTENTIAL FLAW: Possibly copy from a memory location located before the source buffer", lines[line_number-1]) != None or
               re.search("POTENTIAL FLAW: Possibly copy from a memory location located before the source buffer", lines[line_number-3]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "buffer_read", line_number + 1, site_line, None, True))

          #write what where files
          elif(re.search("POTENTIAL FLAW: The following removes 'a' from the list.  Because of the possible overflow this", lines[line_number-16]) != None or
               re.search("POTENTIAL FLAW: The following removes 'a' from the list.  Because of the possible overflow this", lines[line_number-17]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "write_what_where", line_number + 1, site_line, None, True))

          #return files
          elif(re.search("FLAW: returning stack-allocated buffer", lines[line_number-1]) != None):
            site_line = re.sub('(  |\r|\n|\t)', "", line)
            sites.append(site.Site(file_path, "return", line_number + 1, site_line, None, True))

          #elif(re.search("POTENTIAL FLAW: Attempt to access a negative index of the array", lines[line_number-4]) != None):
            #site_line = re.sub('(  |\r|\n|\t)', "", line)
            #sites.append(site.Site(file_path, "buffer_read", line_number + 1, site_line, None, True))

      f.close()
    return sites

  @staticmethod
  def set_buggy_attr(sites, buggy_sites):
    """Iterates through a sites list and set the ones in common with a
    buggy_sites list as buggy.
    """
    for site in sites:
      for buggy_site in buggy_sites:
        # Limitation: some cases have TWO buffer_read sites in one line, where
        # one is buggy (data) and the other is good (strlen(dest)):
        #     memcpy(dest, data, strlen(dest)*sizeof(char));
        # Matching only on line number means the code can't distinguish the two 
        # sites.  The code should match on info (varible), too.  Currently the
        # buggy_site code doesn't extract variable name, so this is not a
        # simple fix.  
        #   There is a special hard-coded check in to_xml() to fix those cases.
        if(buggy_site.filename == site.filename and buggy_site.line == site.line and buggy_site.site_type == site.site_type):
          site.buggy = True

  @staticmethod
  def to_xml(sites, xml_output_path = r'sites_list.xml'):
    """Prints a list of sites to an xml file."""
    files = Set([])
    for site in sites:
      files = files.union(Set([site.filename]))

    root = ET.Element('sites')
    for filename in files:
      file_element = ET.SubElement(root, 'file')
      file_element.set('name', os.path.basename(filename))
      for site in sites:
        if(site.filename == filename 
          # Exclude sites with a wide string literal PEB 25 Sep 2015
          and 'L"%s"' not in site.code
          # Exclude printWLine() sites in certain files.  PrintWLine() uses wide
          # string specifier, "%ls", which is not supported. PEB 13 Oct 2015
          and not (site.site_type == 'buffer_read' and
	('CWE126_Buffer_Overread__CWE170_wchar_t_memcpy_' in site.filename
	 or 'CWE126_Buffer_Overread__CWE170_wchar_t_strncpy_' in site.filename
	 or 'CWE126_Buffer_Overread__CWE170_wchar_t_loop_' in site.filename)
                   and 'printWLine' in site.code)
          ):
          site_element = ET.SubElement(file_element, 'site')
          site_element.set('type', site.site_type)
          site_element.set('line', str(site.line))
          site_element.set('info', site.info)
          site_element.set('val', ("buggy" if (site.buggy 
          # buffer_read sites with strlen(dest) are NOT buggy, but 
          # set_buggy_attr() cannot distinguish.  See comment above in
          # set_buggy_attr(). PEB Oct 2015
          and not (site.site_type == 'buffer_read' and site.info == 'dest' and 'strlen(dest)' in site.code)
) else "good"))
          code_element = ET.SubElement(site_element, 'code')
          code_element.text = ET.CDATA(site.code)

    f = open(xml_output_path, 'w')
    f.write(ET.tostring(root, pretty_print=True, xml_declaration=True, encoding='utf-8'))
    f.close()

  def extract_ast(self):
    """Fills ast_queue with one AST for each file in the filepath."""
    for file_path in self.files:
      ast = parse_file(file_path, use_cpp=True,
              cpp_path=CPPPATH,
              cpp_args=[r'-Iutils/fake_libc_include', r'-iquoteutils/testcasesupport', r'-D_WIN32'],
              )
      self.ast_queue.put(ast)
    self.ast_queue_done = True

  def populate_ast_attributes(self):
    """Calls all the methods to populate each AST list for each site type."""
    while(not (self.ast_queue.empty() and self.ast_queue_done)):
      ast = self.ast_queue.get()
      dbz_th = threading.Thread(None, self.populate_ast_divisions(ast))
      variable_th = threading.Thread(None, self.populate_ast_variable_accesses(ast))
      null_ptr_th = threading.Thread(None, self.populate_ast_null_ptr_dereferences(ast))
      int_of_th = threading.Thread(None, self.populate_ast_int_overflows(ast))
      int_uf_th = threading.Thread(None, self.populate_ast_int_underflows(ast))
      buffer_write_th = threading.Thread(None, self.populate_ast_buffer_writes(ast))
      buffer_read_th = threading.Thread(None, self.populate_ast_buffer_reads(ast))
      www_th = threading.Thread(None, self.populate_ast_write_what_wheres(ast))
      ret_th = threading.Thread(None, self.populate_ast_returns(ast))
      threads = []
      if(self.parse_single_cwe == 'division_by_zero' or not self.parse_single_cwe):
        threads.append(dbz_th)

      if(self.parse_single_cwe == 'variable_access' or not self.parse_single_cwe):
        threads.append(variable_th)

      if(self.parse_single_cwe == 'null_ptr' or not self.parse_single_cwe):
        threads.append(null_ptr_th)

      if(self.parse_single_cwe == 'int_of' or not self.parse_single_cwe):
        threads.append(int_of_th)

      if(self.parse_single_cwe == 'int_uf' or not self.parse_single_cwe):
        threads.append(int_uf_th)

      if(self.parse_single_cwe == 'buffer_write' or not self.parse_single_cwe):
        threads.append(buffer_write_th)

      if(self.parse_single_cwe == 'buffer_read' or not self.parse_single_cwe):
        threads.append(buffer_read_th)

      if(self.parse_single_cwe == 'write_what_where' or not self.parse_single_cwe):
        threads.append(www_th)

      if(self.parse_single_cwe == 'return' or not self.parse_single_cwe):
        threads.append(ret_th)

      for thread in threads:
        thread.start()

      for thread in threads:
        thread.join()

  def populate_ast_divisions(self, ast):
    """Calls pycparser node visitor in the ast. puts the division nodes in the
    ast_divisions queue.
    """
    dbz_visitor = div_by_zero.DivByZeroVisitor()
    dbz_visitor.visit(ast)
    for node in dbz_visitor.nodes:
      self.ast_divisions.put(node)

  def populate_ast_variable_accesses(self, ast):
    """Calls pycparser node visitor in the ast. puts the variable access nodes
    in the ast_variable_access queue.
    """
    variable_access_visitor = variable_access.VariableAccessVisitor()
    variable_access_visitor.visit(ast)
    for node in variable_access_visitor.nodes:
      self.ast_variable_accesses.put(node)

  def populate_ast_null_ptr_dereferences(self, ast):
    """Calls pycparser node visitor in the ast. puts the null pointer
    dereferences nodes in the ast_variable_access queue.
    """
    null_ptr_deref_visitor = null_ptr_deref.NullPtrDerefVisitor()
    null_ptr_deref_visitor.visit(ast)
    for node in null_ptr_deref_visitor.nodes:
      self.ast_null_ptr_dereferences.put(node)

  def populate_ast_int_overflows(self, ast):
    """Calls pycparser node visitor in the ast. puts the integer overflow
    nodes in the ast_int_overflow queue.
    """
    int_overflow_visitor = int_overflow.IntOverflowVisitor()
    int_overflow_visitor.visit(ast)
    for node in int_overflow_visitor.nodes:
      self.ast_int_overflows.put(node)

  def populate_ast_int_underflows(self, ast):
    """Calls pycparser node visitor in the ast. puts the integer underflow
    nodes in the ast_int_underflow queue.
    """
    int_underflow_visitor = int_underflow.IntUnderflowVisitor()
    int_underflow_visitor.visit(ast)
    for node in int_underflow_visitor.nodes:
      self.ast_int_underflows.put(node)

  def populate_ast_buffer_writes(self, ast):
    """Calls pycparser node visitor in the ast. puts the buffer write
    nodes in the ast_buffer_writes queue.
    """
    buffer_write_visitor = buffer_write.BufferWriteVisitor()
    buffer_write_visitor.visit(ast)
    for node in buffer_write_visitor.nodes:
      self.ast_buffer_writes.put(node)

  def populate_ast_buffer_reads(self, ast):
    """Calls pycparser node visitor in the ast. puts the buffer read
    nodes in the ast_buffer_reads queue.
    """
    buffer_read_visitor = buffer_read.BufferReadVisitor()
    buffer_read_visitor.visit(ast)
    for node in buffer_read_visitor.nodes:
      self.ast_buffer_reads.put(node)

  def populate_ast_write_what_wheres(self, ast):
    """Calls pycparser node visitor in the ast. puts the buffer read
    nodes in the ast_buffer_reads queue.
    """
    write_what_where_visitor = write_what_where.WriteWhatWhereVisitor()
    write_what_where_visitor.visit(ast)
    for node in write_what_where_visitor.nodes:
      self.ast_write_what_wheres.put(node)

  def populate_ast_returns(self, ast):
    """Calls pycparser node visitor in the ast. puts the buffer read
    nodes in the ast_buffer_reads queue.
    """
    return_visitor = returned.ReturnVisitor()
    return_visitor.visit(ast)
    for node in return_visitor.nodes:
      self.ast_returns.put(node)

# end of extractor.py
